// Create candidate 3-CNF-SAT puzzles for Chicken Shock. Use a simple
// recursive search and count the depth of the game tree as a proxy for
// its difficulty.

// Copyright 2023 Kristofer Munsterhjelm.
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of  MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// this program.  If not, see <http://www.gnu.org/licenses/>.

#include <algorithm>
#include <array>
#include <cassert>
#include <iostream>
#include <iterator>
#include <numeric>
#include <vector>

// Each clause consists of three literals. A literal is a variable number
// along with a boolean that is true if it's counted directly, and false if
// its negation is counted.

// The variable indices are:
//	0	blue
//	1	green
//	2	cyan
//	3	red
//	4	purple
//	5	yellow

// The optional seventh color is grey, which can be created in ZZT by putting
// sharks and then either solids or fakes on top. But I ended up not using it
// in Chicken Shock as six colors turned out to be plenty.

enum color {BLUE = 0, GREEN = 1, CYAN = 2, RED = 3, PURPLE = 4, YELLOW = 5,
	GREY = 6};

const int FULL_COLORS = 6; // Not counting grey which is kind of hacky
const int ALL_COLORS = 7; // Counting it.

// There are eight clauses, but I may extend this.

// There's also a dummy white switch, but it's only for trolling purposes.

// If this is too easy, I might extend to:
//  6   grey (only the first element)
// with objects on the white solids flipping them from grey fake to grey solid
// by using sharks.

typedef std::pair<int, bool> literal;
typedef std::array<literal, 3> clause;

// literal bool		evaluated		result
//	true			true			true
//	true			false			false
//	false			true			false
//	false			false			true

// I.e. the result is true iff the two are equal.

bool evaluate(const literal & lit_to_check,
	const std::vector<bool> & var_assignment) {

	return var_assignment[lit_to_check.first] ==
		lit_to_check.second;
}

bool satisfiable(const clause & check, 
	const std::vector<bool> & var_assignment) {

	return evaluate(check[0], var_assignment) ||
		evaluate(check[1], var_assignment) ||
		evaluate(check[2], var_assignment);
}

void print_boolean_parameters(std::vector<bool> & parameters) {

	for (bool b: parameters) {
		if (b) {
			std::cout << "# ";
		} else {
			std::cout << ". ";
		}
	}
}

void print_boolean_parameters(std::vector<bool> & parameters,
	std::vector<bool> & has_assigned) {

	for (size_t i = 0; i < parameters.size(); ++i) {
		if (!has_assigned[i]) {
			std::cout << "? ";
			continue;
		}
		bool b = parameters[i];
		if (b) {
			std::cout << "# ";
		} else {
			std::cout << ". ";
		}
	}
}



bool has_solution(std::vector<clause> & clauses,
	std::vector<bool> & assignment, std::vector<bool> & has_been_assigned,
	int clause_level, int literal_level, int & nodes_visited, bool verbose) {

	// If clause_level is greater than the number of clauses, return true.
	// Otherwise, check the literal given by clause_level and literal_level.
	// If it's true, go to the next clause. If it's false, go to the next
	// literal unless we're at the last one (in which case return false).
	// If we haven't assigned a value to the variable that's being checked
	// by the literal, recurse twice: once for true and once for false.

	if (clause_level >= clauses.size()) {
		if (verbose) {
			std::cout << "Solution: ";
			print_boolean_parameters(assignment, has_been_assigned);
			std::cout << "\n";
		}
	 return true; }
	if (literal_level == 3) {
		return has_solution(clauses, assignment, has_been_assigned,
			clause_level+1, 0, nodes_visited, verbose);
	}

	++nodes_visited;

	literal cur_literal = clauses[clause_level][literal_level];

	if (!has_been_assigned[cur_literal.first]) {
		// Make the recursion unbiased by testing the hypotheses
		// in random order.
		bool first_try = drand48() > 0.5;

		has_been_assigned[cur_literal.first] = true;
		assignment[cur_literal.first] = first_try;
		if (has_solution(clauses, assignment, has_been_assigned,
			clause_level, literal_level, nodes_visited, verbose)) {
			has_been_assigned[cur_literal.first] = false;
			return true;
		}
		assignment[cur_literal.first] = !first_try;
		// If neither setting the variable to true nor to false works,
		// then the problem is unsatisfiable given the variables assigned
		// so far.
		bool outcome = has_solution(clauses, assignment, has_been_assigned,
			clause_level, literal_level, nodes_visited, verbose);
		has_been_assigned[cur_literal.first] = false;
		return outcome;
	}

	if (evaluate(cur_literal, assignment)) {
		return has_solution(clauses, assignment, has_been_assigned,
			clause_level+1, 0, nodes_visited, verbose);
	} else {
		// We can't have recursed from an earlier literal being true,
		// so if literal level is and evaluate is false, the whole
		// clause must be false, which means the current solution isn't
		// right.
		if (literal_level == 2) {
			return false;
		}
		return has_solution(clauses, assignment, has_been_assigned,
			clause_level, literal_level + 1, nodes_visited, verbose);
	}
}

// Get the minimum distance from either all true or all false.
int hamming_error(const std::vector<bool> & to_check,
	std::vector<bool> & has_been_assigned) {

	int trues = 0, falses = 0;

	for (size_t i = 0; i < to_check.size(); ++i) {
		if (!has_been_assigned[i]) { continue; }
		if (to_check[i]) {
			++trues;
		} else {
			++falses;
		}
	}

	// Trues is the Hamming distance from all-false,
	// and falses is the Hamming distance from all-true.
	return std::min(trues, falses);
}

int hamming_error(const std::vector<bool> & to_check) {

	int trues = 0, falses = 0;

	for (size_t i = 0; i < to_check.size(); ++i) {
		if (to_check[i]) {
			++trues;
		} else {
			++falses;
		}
	}

	// Trues is the Hamming distance from all-false,
	// and falses is the Hamming distance from all-true.
	return std::min(trues, falses);
}


void get_hamming_stats(std::vector<clause> & clauses,
	std::vector<bool> & assignment, std::vector<bool> & has_been_assigned,
	std::vector<int> & hamming_stats, int clause_level, int literal_level) {

	if (clause_level >= clauses.size()) {
		/*std::cout << "Get hamming stats: Solution: ";
		print_boolean_parameters(assignment, has_been_assigned);
		std::cout << std::endl;*/
		hamming_stats.push_back(hamming_error(assignment, has_been_assigned));
		return;
	}
	if (literal_level == 3) {
		return get_hamming_stats(clauses, assignment, has_been_assigned,
			hamming_stats, clause_level+1, 0);
	}

	literal cur_literal = clauses[clause_level][literal_level];

	if (!has_been_assigned[cur_literal.first]) {
		//std::cout << "(" << clause_level << ", " << literal_level << "): Checking literal color" << cur_literal.first << " = true " << std::endl;
		has_been_assigned[cur_literal.first] = true;
		assignment[cur_literal.first] = true;
		get_hamming_stats(clauses, assignment, has_been_assigned,
			hamming_stats, clause_level, literal_level);

		//std::cout << "(" << clause_level << ", " << literal_level << "): Checking literal color" << cur_literal.first << " = false " << std::endl;
		assignment[cur_literal.first] = false;
		get_hamming_stats(clauses, assignment, has_been_assigned,
			hamming_stats, clause_level, literal_level);
		has_been_assigned[cur_literal.first] = false;
		return;
	}

	if (evaluate(cur_literal, assignment)) {
		//std::cout << "(" << clause_level << ", " << literal_level << "): Literal OK" << std::endl;
		return get_hamming_stats(clauses, assignment, has_been_assigned,
			hamming_stats, clause_level+1, 0);
	} else {
		if (literal_level == 2) {
			//std::cout << "(" << clause_level << ", " << literal_level << "): Literal not OK, exhausted all literals" << std::endl;
			return;
		}
		//std::cout << "(" << clause_level << ", " << literal_level << "): Literal not OK, trying next" << std::endl;
		return get_hamming_stats(clauses, assignment, has_been_assigned,
			hamming_stats, clause_level, literal_level + 1);
	}
}

bool has_solution(std::vector<clause> & clauses, int num_vars,
	int & out_nodes_visited, bool verbose) {

	std::vector<bool> assignment(num_vars, false),
		has_been_assigned(num_vars, false);

	return has_solution(clauses, assignment, has_been_assigned,
		0, 0, out_nodes_visited, verbose);
}

std::vector<int> get_stats(std::vector<clause> & clauses, int num_vars) {
	std::vector<bool> assignment(num_vars, false),
		has_been_assigned(num_vars, false);

	std::vector<int> hamming;

	get_hamming_stats(clauses, assignment, has_been_assigned,
		hamming, 0, 0);	

	return hamming;
}

int min_int(const std::vector<int> & x) {
	return *std::min_element(x.begin(), x.end());
}

void print_stats(unsigned long long puzzle_ID, int node_visits, std::vector<int> & hamming) {

	std::cout << puzzle_ID << ": Hamming info: ";
	std::sort(hamming.begin(), hamming.end());
	std::copy(hamming.begin(), hamming.end(), std::ostream_iterator<int>(std::cout, " "));

	double minimum = *std::min_element(hamming.begin(), hamming.end());
	double mean = std::accumulate(hamming.begin(), hamming.end(), 0.0)/hamming.size();

	std::cout << "\n" << puzzle_ID << ": node visits: " << node_visits
		<< " solutions: " << hamming.size() << ", min flips: "
		<< minimum << " mean flips: " << mean << std::endl;

}

//	0	blue
//	1	green
//	2	cyan
//	3	red
//	4	purple
//	5	yellow


void print_puzzle(const std::vector<clause> & clauses) {
	std::vector<std::string> true_colors = 
		{"blue", "green", "cyan", "red", "purple", "yellow", "grey"};
	std::vector<std::string> false_colors =
		{"!BLUE", "!GREEN", "!CYAN", "!RED", "!PURPLE", "!YELLOW", "!GREY"};

	for (const clause & cur_clause: clauses) {
		std::cout << "\t[ ";
		for (const literal & cur_literal: cur_clause) {
			if (cur_literal.second) {
				std::cout << true_colors[cur_literal.first] << " ";
			} else {
				std::cout << false_colors[cur_literal.first] << " ";
			}
		}
		std::cout << "]\n";
	}
}

void brute_force(const std::vector<clause> & clauses,
	std::vector<bool> & assignment,
	int variable_level,
	std::vector<int> & out_hamming_stats) {

	if (variable_level == assignment.size()) {
		for (const clause & c: clauses) {
			if (!satisfiable(c,  assignment)) {
				return;
			}
		}

		// Dump the Hamming stats result.
		out_hamming_stats.push_back(hamming_error(assignment));
		/*std::cout << "Brute force: Solution: ";
		print_boolean_parameters(assignment);
		std::cout << " hamming " << hamming_error(assignment) << std::endl;*/
		return;
	}

	assignment[variable_level] = false;
	brute_force(clauses, assignment, variable_level+1, out_hamming_stats);
	assignment[variable_level] = true;
	brute_force(clauses, assignment, variable_level+1, out_hamming_stats);
}

std::vector<int> brute_force_hamming(const std::vector<clause> & clauses,
	int num_variables){

	std::vector<int> hamming_out;
	std::vector<bool> assignment(num_variables, true);

	brute_force(clauses, assignment, 0, hamming_out);

	return hamming_out;
}
	
int main() {
	const int MAX_NUM_CLAUSES = 11,
			  NUM_VARIABLES = FULL_COLORS; // or ALL_COLORS depending

	// The following ints control the best case and median difficulty
	// of the puzzle. The best case roughly corresponds to how easy it
	// is to happen upon a solution by being lucky; the median to how
	// easy it is to solve it by thinking.

	const int min_difficulty = 30,
			  med_difficulty = 75;

	// Set to true to get solutions. But beware: this can make puzzles
	// seem a lot easier than they really are.
	bool verbose = false;

	const unsigned long long ITERS = 40e9;

	std::vector<clause> clauses;

	// Set this to 28865581800 to recreate the puzzle used in
	// Chicken Shock.
	unsigned long long start = 0;

	std::cout << "Starting the generator. This may take a long time.\n";

	for (unsigned long long glob = start; glob < ITERS; ++glob) {

		unsigned int seed = glob/10;

		if (glob % 10 == 0) {
			srand(seed);
			srand48(seed);
			srandom(seed);
		}

		clauses.clear();

		int clause_ceiling = 2 + random() % (MAX_NUM_CLAUSES - 2);

		for (int i = 0; i < 100 && clauses.size() < clause_ceiling; ++i) {
			clause clause_to_add;
			std::vector<bool> picked(ALL_COLORS, false);
			for (int j = 0; j < 3; ++j) {
				if (j != 0) {
					picked[GREY] = true; // grey is only allowed in the first slot.
				}

				bool wrong_adjacency;

				literal literal_to_add;
				do {
					wrong_adjacency = false;
					literal_to_add.first = random() % NUM_VARIABLES;
					// Placing purple next to green is not allowed,
					// nor is placing green next to cyan, because
					// they're too hard to tell apart.
					if (j != 0) {
						color prev = (color)clause_to_add[j-1].first,
							now = (color)literal_to_add.first;

						wrong_adjacency =
							(now == GREEN && prev == CYAN)
							|| (now == CYAN && prev == GREEN)
							|| (now == PURPLE && prev == RED)
							|| (now == RED && prev == PURPLE);
					}
				} while (picked[literal_to_add.first] || wrong_adjacency);

				// Comment the next line out to get dupes, 
				// e.g. "RED GREEN RED" type doors.
				picked[literal_to_add.first] = true;
				literal_to_add.second = drand48() > 0.5;
				clause_to_add[j] = literal_to_add;
			}
			clauses.push_back(clause_to_add);
			int throwaway = 0;
			if (!has_solution(clauses, NUM_VARIABLES, throwaway, false)) {
				clauses.pop_back();
				// If we have no solution, also remove one clause at
				// random. Seems to give slightly better results? 
				if (clauses.size() > 3 && drand48() > 0.5) {
					std::random_shuffle(clauses.begin(), clauses.end());
					clauses.pop_back();
				}
			}
		}

		int nodes_visited = 0;
		if (has_solution(clauses, NUM_VARIABLES, nodes_visited, false)) {
			//if (nodes_visited < 10) { continue;}

			std::vector<int> hamming = get_stats(clauses, NUM_VARIABLES);

			if (min_int(hamming) < 3) { continue; }

			// The current hamming vector may not be counting properly because
			// its search sometimes includes wildcards (variables that can be
			// either true or false). So use a brute force search to get the
			// actual distribution.
			hamming = brute_force_hamming(clauses, NUM_VARIABLES);

			// Get the median visit count for 15 random rearrangements of the
			// puzzle.

			std::vector<int> rearr_node_visits;
			std::vector<clause> clauses_rearr = clauses;

			for (int j = 0; j < 15; ++j) {
				int nodes_visited_inner = 0;
				std::random_shuffle(clauses_rearr.begin(), clauses_rearr.end());
				has_solution(clauses_rearr, NUM_VARIABLES, nodes_visited_inner, false);
				rearr_node_visits.push_back(nodes_visited_inner);
			}

			std::sort(rearr_node_visits.begin(), rearr_node_visits.end());

			int min_visits = rearr_node_visits[0],
				median_visits = rearr_node_visits[7];

			if (median_visits < med_difficulty 
				&& min_visits < min_difficulty) { continue; }

			std::cout << "----------------------------------------\n";
			std::cout << glob << ": satisfiable in " << median_visits << "\n";
			print_puzzle(clauses);
			print_stats(glob, median_visits, hamming);
			has_solution(clauses_rearr, NUM_VARIABLES, nodes_visited, verbose);
			std::cout << glob << ": node visits: min: " 
				<< rearr_node_visits[0] << " med: " << 
				median_visits << " max: " << *rearr_node_visits.rbegin() << std::endl;
		}
	}
}